/* Hanoi   -- Solve the ancient problem of the Hanoi Towers. Demo solution */
/*            included.                                                    */
/*                                                                         */
/* Copyright (c) 1996 Leonidas Vomvoridis    a95-3597@amalia.ee.upatras.gr */
/*                                                                         */
/* This program is free software; you can redistribute it and/or modify it */
/* under the terms of the  GNU General Public License  as published by the */
/* Free Software Foundation.  This program is distributed in the hope that */
/* it will be useful, but  WITHOUT ANY WARRANTY; without even the  implied */
/* warranty of  MERCHANTABILITY  or  FITNESS FOR A PARTICULAR PURPOSE. See */
/* the GNU General Public License for more details.                        */



#include<stdio.h>
#include<string.h>
#include"hanoi.h"


char PIECES=5;                  /* default no of pieces */
char towers[3][MAXPIECES];      /* piece layout matrix  */


int main()
{
    char choice,new_pieces;
    printf("\t\tThe Towers Of Hanoi (c) 1996 Leonidas Vomvoridis");
    init();
    for(;;)
    {
       display_menu();
       scanf("%d",&choice);
       /* Start menu option control */
       if(choice==1) /* Intaractive game */
       {
          interact();
          continue;
       }
       if(choice==2) /* Demo solution */
       {
          demo();
          continue;
       }
       if(choice==3) /* Change number of PIECES */
       {
          printf("\n\n\nNumber of pieces is currently %d.\n",PIECES);
          printf("Enter new value (1-%d): ",MAXPIECES);
          scanf("%d",&new_pieces);
          if((new_pieces>=1)&&(new_pieces<=MAXPIECES))
             PIECES=new_pieces;
          else printf("Invalid number. Sorry.\n");
          init();
          continue;
       }
       if(choice==4) /* Exit */
       {
          printf("Exiting to %s.\n",OSTYPE);
          break;
       }

       printf("\nInvalid choice. Only 1 thru 4 please.\n");
       /* End menu option control */
    }
    return 0;
}



init() /* Initialize matrix */
{
    char i;
    for(i=(PIECES-1);i>=0;i--)
    {
       towers[0][i]=i+1;
       towers[1][i]=towers[2][i]=0;
    }
}



interact()    /* Playable game interface */
{
    char a,b;
    char cmd[3]; /* Bug: if cmd is longer ??? */
    printf("\nStarting playable game with %d pieces",PIECES);
    display();
    for(;;)
    {  /* Start command control */
       printf("Command (? for help): ");
       scanf("%s",cmd);
       if(cmd[0]=='q') return;
       if(cmd[0]=='?')
       {
          help();
          continue;
       }
       if(cmd[0]=='n')
       {
          init();
          display();
          continue;
       }
       a=cmd[0]-'0';
       b=cmd[2]-'0';
       if(a>=1&&a<=3&&b>=1&&b<=3) /* choice is valid */
       {
          if(!(move((a-1),(b-1)))) display();    /* move() returned 0 */
          else printf("Illegal move, %s\n",cmd); /* move() returned 1 */
       }
       else printf("Invalid command, %s\n",cmd);
    }  /* End command control */
}



demo()   /* Demo solution */
{
    init();
    printf("Solving %d-piece game.",PIECES);
    printf(" %d moves will be required.",((1<<PIECES)-1));
    display();
    presskey();
    solve(PIECES,0,2); /* move all from 1 to 3 */
}



/* Solution algorithm */
solve(char n,char from,char to)
{
    char alt;
    alt=3-(from+to);  /* alternate col */
    if(n==1)
    {
        printf("Moving %d-%d...",(from+1),(to+1));
        move(from,to);
        display();
        presskey();
    }
    else
    {
        solve((n-1),from,alt);
        printf("Moving %d-%d...",(from+1),(to+1));
        move(from,to);
        display();
        presskey();
        solve((n-1),alt,to);
    }
}



char move(char from,char to)   /* Controls if move is possible, moves the */
{                              /* topmost piece from tower a to tower b.  */
    char i,j;
    char val_a=0,val_b=0;
    i=j=(PIECES-1);
    while(towers[from][i]!=0&&i>=0) val_a=towers[from][i--];
    while(towers[to][j]!=0&&i<PIECES) val_b=towers[to][j--];
    if(val_a!=0&&(val_b==0||val_a<val_b))           /* move is legal */
    {
      towers[to][j]=val_a;
      towers[from][(i+1)]=0;
      return 0;                /* ok */
    }
    else return 1;             /* Error */
}



display()               /* Displays ascii layout of the three towers. */
{                       /* MAXPIECES=11 for a 80x25 display (default) */
    char i,j,k;
    printf("\n\n\n\n\n\n");
    for(i=0;i<=(MAXPIECES-PIECES);i++) putchar('\n');
    for(j=0;j<PIECES;j++)              /* sweep rows */
    {
       for(i=0;i<(35-3*PIECES);i++) putchar(' '); /* center towers (80x25) */
       for(i=0;i<3;i++)                /* sweep columns */
          for(k=0;k<(2*PIECES+3);k++)  /* sweep piece width */
          {  /* Draw ascii display of each piece */
             if(k<(PIECES+1-towers[i][j])) putchar(' ');
             if(k>=(PIECES+1-towers[i][j])&&k<(PIECES+1)) putchar('=');
             if(k==(PIECES+1)) putchar('I');
             if(k>(PIECES+1)&&k<=(PIECES+1+towers[i][j])) putchar('=');
             if(k>(PIECES+1+towers[i][j])) putchar(' ');
          }
       putchar('\n');
    }
    for(i=0;i<(35-3*PIECES);i++) putchar(' '); /* center towers (80x25 res) */
    for(i=0;i<(6*PIECES+9);i++) putchar('#');  /* draw base */
    printf("\n\n\n\n\n\n");
}



display_menu()
{
    printf("\n\n\n\n\n\n\n\n\n\n\t\tMain Menu:\n\n");
    printf("\t\t1. Interactive Game\n");
    printf("\t\t2. Demo Solution\n");
    printf("\t\t3. Change number of pieces (1-%d, now: %d)\n",MAXPIECES,PIECES);
    printf("\t\t4. Exit to %s\n\n\n\n\n\n\n\n\n",OSTYPE);
    printf("\t\tYour choice:");
}



help()
{
    printf("Valid commands:\n");
    printf("?  \t\tthis help screen\n");
    printf("n  \t\tnew game\n");
    printf("q  \t\tquit\n");
    printf("a-b\t\tmove topmost piece of tower a to tower b.\n");
    printf("Tower a must _not_ be empty. If tower b is not empty, the\n");
    printf("topmost piece of b _must_ be larger than the one you are\n");
    printf("moving from a. Please use a dash between the two numbers.\n");
    printf("Both a and b must be between one of 1, 2 and 3, and a must\n");
    printf("obviously be different than b.\n\n");
}



presskey()
{
    printf("Press enter to coninue...");
    fflush(stdin); /* To ignore \n from user's previous command */
    getchar();
    fflush(stdin); /* To ignore accidental keypresses */
    putchar('\n');
}

